# -*- coding: utf-8 -*-
"""
Database models for Enterprise Integrated Channel SAP SuccessFactors.
"""

from __future__ import absolute_import, unicode_literals

import json
from logging import getLogger

from config_models.models import ConfigurationModel
from integrated_channels.integrated_channel.models import EnterpriseCustomerPluginConfiguration
from integrated_channels.sap_success_factors.exporters.course_metadata import SapSuccessFactorsCourseExporter
from integrated_channels.sap_success_factors.exporters.learner_data import SapSuccessFactorsLearnerExporter
from integrated_channels.sap_success_factors.transmitters.course_metadata import SapSuccessFactorsCourseTransmitter
from integrated_channels.sap_success_factors.transmitters.learner_data import SapSuccessFactorsLearnerTransmitter
from simple_history.models import HistoricalRecords

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

LOGGER = getLogger(__name__)


@python_2_unicode_compatible
class SAPSuccessFactorsGlobalConfiguration(ConfigurationModel):
    """
    The global configuration for integrating with SuccessFactors.
    """

    completion_status_api_path = models.CharField(max_length=255)
    course_api_path = models.CharField(max_length=255)
    oauth_api_path = models.CharField(max_length=255)
    provider_id = models.CharField(max_length=100, default='EDX')

    class Meta:
        app_label = 'sap_success_factors'

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return "<SAPSuccessFactorsGlobalConfiguration with id {id}>".format(id=self.id)

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()


@python_2_unicode_compatible
class SAPSuccessFactorsEnterpriseCustomerConfiguration(EnterpriseCustomerPluginConfiguration):
    """
    The Enterprise-specific configuration we need for integrating with SuccessFactors.
    """

    USER_TYPE_USER = 'user'
    USER_TYPE_ADMIN = 'admin'

    USER_TYPE_CHOICES = (
        (USER_TYPE_USER, 'User'),
        (USER_TYPE_ADMIN, 'Admin'),
    )

    key = models.CharField(
        max_length=255,
        verbose_name="Client ID",
        help_text=_("OAuth client identifier.")
    )
    sapsf_base_url = models.CharField(
        max_length=255,
        verbose_name="SAP Base URL",
        help_text=_("Base URL of success factors API.")
    )
    sapsf_company_id = models.CharField(
        max_length=255, verbose_name="SAP Company ID", help_text=_("Success factors company identifier.")
    )
    sapsf_user_id = models.CharField(
        max_length=255,
        verbose_name="SAP User ID",
        help_text=_("Success factors user identifier.")
    )
    secret = models.CharField(
        max_length=255,
        verbose_name="Client Secret",
        help_text=_("OAuth client secret.")
    )
    user_type = models.CharField(
        max_length=20,
        choices=USER_TYPE_CHOICES,
        default=USER_TYPE_USER,
        verbose_name="SAP User Type",
        help_text=_("Type of SAP User (admin or user).")
    )

    history = HistoricalRecords()

    class Meta:
        app_label = 'sap_success_factors'

    def __str__(self):
        """
        Return human-readable string representation.
        """
        return "<SAPSuccessFactorsEnterpriseCustomerConfiguration for Enterprise {enterprise_name}>".format(
            enterprise_name=self.enterprise_customer.name
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()

    @staticmethod
    def channel_code():
        """
        Returns an capitalized identifier for this channel class, unique among subclasses.
        """
        return 'SAP'

    @property
    def provider_id(self):
        '''
        Fetch ``provider_id`` from global configuration settings
        '''
        return SAPSuccessFactorsGlobalConfiguration.current().provider_id

    def get_learner_data_transmitter(self):
        """
        Return a ``SapSuccessFactorsLearnerTransmitter`` instance.
        """
        return SapSuccessFactorsLearnerTransmitter(self)

    def get_learner_data_exporter(self, user):
        """
        Return a ``SapSuccessFactorsLearnerDataExporter`` instance.
        """
        return SapSuccessFactorsLearnerExporter(user, self)

    def get_course_data_transmitter(self):
        """
        Return a ``SapSuccessFactorsCourseTransmitter`` instance.
        """
        return SapSuccessFactorsCourseTransmitter(self)

    def get_course_data_exporter(self, user):
        """
        Return a ``SapSuccessFactorsCourseExporter`` instance.
        """
        return SapSuccessFactorsCourseExporter(user, self)


@python_2_unicode_compatible
class SapSuccessFactorsLearnerDataTransmissionAudit(models.Model):
    """
    The payload we sent to SuccessFactors at a given point in time for an enterprise course enrollment.
    """

    sapsf_user_id = models.CharField(max_length=255, blank=False, null=False)
    enterprise_course_enrollment_id = models.PositiveIntegerField(blank=False, null=False)
    course_id = models.CharField(max_length=255, blank=False, null=False)
    course_completed = models.BooleanField(default=True)
    instructor_name = models.CharField(max_length=255, blank=True)
    grade = models.CharField(max_length=100, blank=False, null=False)

    # We send a UNIX timestamp to SAPSF.
    completed_timestamp = models.BigIntegerField()

    # Request-related information.
    status = models.CharField(max_length=100, blank=False, null=False)
    error_message = models.TextField(blank=True)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        app_label = 'sap_success_factors'

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return (
            '<SapSuccessFactorsLearnerDataTransmissionAudit {transmission_id} for enterprise enrollment '
            '{enterprise_course_enrollment_id}, SAPSF user {sapsf_user_id}, and course {course_id}>'.format(
                transmission_id=self.id,
                enterprise_course_enrollment_id=self.enterprise_course_enrollment_id,
                sapsf_user_id=self.sapsf_user_id,
                course_id=self.course_id
            )
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()

    @property
    def provider_id(self):
        """
        Fetch ``provider_id`` from global configuration settings
        """
        return SAPSuccessFactorsGlobalConfiguration.current().provider_id

    def serialize(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Return a JSON-serialized representation.

        Sort the keys so the result is consistent and testable.

        # TODO: When we refactor to use a serialization flow consistent with how course metadata
        # is serialized, remove the serialization here and make the learner data exporter handle the work.
        """
        return json.dumps(self._payload_data(), sort_keys=True)

    def _payload_data(self):
        """
        Convert the audit record's fields into SAP SuccessFactors key/value pairs.
        """
        return dict(
            userID=self.sapsf_user_id,
            courseID=self.course_id,
            providerID=self.provider_id,
            courseCompleted="true" if self.course_completed else "false",
            completedTimestamp=self.completed_timestamp,
            grade=self.grade,
        )



# -*- coding: utf-8 -*-
"""
Assist integrated channels with retrieving course metadata.

Module contains resources for integrated pipelines to retrieve all the
metadata for courses in the course catalog belonging to a particular
enterprise customer.
"""

from __future__ import absolute_import, unicode_literals

import json
from logging import getLogger

from integrated_channels.integrated_channel.exporters import Exporter

from enterprise.api_client.enterprise import EnterpriseApiClient
from enterprise.api_client.lms import parse_lms_api_datetime

LOGGER = getLogger(__name__)


class CourseExporter(Exporter):
    """
    Base class for course metadata exporters.
    """

    AVAILABILITY_CURRENT = 'Current'
    AVAILABILITY_UPCOMING = 'Upcoming'
    AVAILABILITY_ARCHIVED = 'Archived'

    def __init__(self, user, enterprise_configuration):
        """
        Instantiate the following variables for use by subclassing course metadata exporters:

        removed_courses_resolved (boolean): Indicates whether `resolve_removed_courses` has been called yet.
        courses (list): A list of courses to be exported.
        """
        super(CourseExporter, self).__init__(user, enterprise_configuration)
        self.removed_courses_resolved = False
        self.courses = []
        client = EnterpriseApiClient(self.user)
        enterprise_course_runs = client.get_enterprise_course_runs(self.enterprise_customer)
        LOGGER.info('Retrieving course run list for enterprise [%s]', self.enterprise_customer.name)
        for course_run in enterprise_course_runs.values():
            self.add_course_run(course_run)

    def export(self):
        """
        Export the set of courses generically by encoding them in JSON format, plus the HTTP method to use.
        """
        yield json.dumps(self.courses, sort_keys=True).encode('utf-8'), 'POST'

    @property
    def data(self):
        """
        Return blanket data by default.

        If a course exporter requires data transformation, it should override this property.
        """
        return {}

    def add_course_run(self, course_run):
        """
        Transform the details of a course run, and save it to our courses list.
        """
        transformed = self.transform(course_run)
        LOGGER.info(
            'Sending course run with plugin configuration [%s]: [%s]',
            self.enterprise_configuration,
            json.dumps(transformed, indent=4),
        )
        self.courses.append(transformed)

    def transform(self, course_run):
        """
        Parse the provided course into the format natively supported by the provider.
        """
        LOGGER.info('Processing course run with ID [%s]', course_run['key'])
        LOGGER.debug('Parsing course run for [%s]: [%s]', self.enterprise_customer, json.dumps(course_run, indent=4))

        # Add the enterprise customer to the course run details so it can be used in the data transform
        course_run['enterprise_customer'] = self.enterprise_customer
        transformed_data = {}
        for key, transform in self.data.items():
            transformed_data[key] = transform(course_run) if transform is not None else course_run.get(key)
        return transformed_data

    def resolve_removed_courses(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Default way to resolve the removal of courses.

        This is implemented as such to allow certain integrated channels that do not really utilize
        this functionality to just return the entire set of ready-to-transmit courses for auditing
        purposes.

        Note that such integrated channels are presumed to be handling updating their course
        availability on upstream catalogs through some other technique, i.e. DELETE requests
        made to the integrated channel's API to remove a course from the upstream catalog.

        An example of an integrated channel that may utilize this: SAPSF -- this function
        would be overridden to generate a course metadata payload that embeds availability
        data to have a correct audit summary at the end of the transmission.
        """
        courses = self.courses if not self.removed_courses_resolved else {}
        self.removed_courses_resolved = True
        return courses

    def format_title(self, course_run):
        """
        Return the transformed version of the course title, as well as the locale. For all instructor-paced courses
        also include the start date to distinguish multiple runs of the same course (ENT-782)
        """
        title = course_run.get('title') or ''
        course_run_start = course_run.get('start')
        if course_run_start and course_run.get('pacing_type') == 'instructor_paced':
            title += ' (Starts: {:%B %Y})'.format(
                parse_lms_api_datetime(course_run_start)
            )
        return title



# -*- coding: utf-8 -*-
"""
URLs for enterprise.
"""
from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.conf.urls import include, url

from enterprise.views import GrantDataSharingPermissions, RouterView

ENTERPRISE_ROUTER = RouterView.as_view()

urlpatterns = [
    url(
        r'^enterprise/grant_data_sharing_permissions',
        GrantDataSharingPermissions.as_view(),
        name='grant_data_sharing_permissions'
    ),
    url(
        r'^enterprise/handle_consent_enrollment/(?P<enterprise_uuid>[^/]+)/course/{}/$'.format(
            settings.COURSE_ID_PATTERN
        ),
        ENTERPRISE_ROUTER,
        name='enterprise_handle_consent_enrollment'
    ),
    url(
        r'^enterprise/(?P<enterprise_uuid>[^/]+)/course/{}/enroll/$'.format(settings.COURSE_ID_PATTERN),
        ENTERPRISE_ROUTER,
        name='enterprise_course_enrollment_page'
    ),
    url(
        r'^enterprise/(?P<enterprise_uuid>[^/]+)/program/(?P<program_uuid>[^/]+)/enroll/$',
        ENTERPRISE_ROUTER,
        name='enterprise_program_enrollment_page'
    ),
    url(
        r'^enterprise/api/',
        include('enterprise.api.urls'),
        name='enterprise_api'
    ),
]

# Because ROOT_URLCONF points here, we are including the urls from the other apps here for now.
urlpatterns += [
    url(
        r'',
        include('consent.urls'),
        name='consent'
    ),
]



# -*- coding: utf-8 -*-
"""
Utilities to get details from the course catalog API.
"""
from __future__ import absolute_import, unicode_literals

import datetime
import logging
from functools import wraps
from time import time

from edx_rest_api_client.client import EdxRestApiClient
from opaque_keys.edx.keys import CourseKey
from requests import Session
from requests.exceptions import ConnectionError, Timeout  # pylint: disable=redefined-builtin
from slumber.exceptions import HttpNotFoundError, SlumberBaseException

from django.conf import settings
from django.utils import timezone

from enterprise.constants import COURSE_MODE_SORT_ORDER
from enterprise.utils import NotConnectedToOpenEdX

try:
    from student.models import CourseEnrollment
except ImportError:
    CourseEnrollment = None

try:
    from openedx.core.djangoapps.embargo import api as embargo_api
except ImportError:
    embargo_api = None

try:
    from openedx.core.lib.token_utils import JwtBuilder
except ImportError:
    JwtBuilder = None


LOGGER = logging.getLogger(__name__)
LMS_API_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
LMS_API_DATETIME_FORMAT_WITHOUT_TIMEZONE = '%Y-%m-%dT%H:%M:%S'


class LmsApiClient(object):
    """
    Object builds an API client to make calls to the edxapp LMS API.

    Authenticates using settings.EDX_API_KEY.
    """

    API_BASE_URL = settings.LMS_INTERNAL_ROOT_URL + '/api/'
    APPEND_SLASH = False

    def __init__(self):
        """
        Create an LMS API client, authenticated with the API token from Django settings.
        """
        session = Session()
        session.headers = {"X-Edx-Api-Key": settings.EDX_API_KEY}
        self.client = EdxRestApiClient(
            self.API_BASE_URL, append_slash=self.APPEND_SLASH, session=session
        )


class JwtLmsApiClient(object):
    """
    LMS client authenticates using a JSON Web Token (JWT) for the given user.
    """

    API_BASE_URL = settings.LMS_INTERNAL_ROOT_URL + '/api/'
    APPEND_SLASH = False

    def __init__(self, user, expires_in=settings.OAUTH_ID_TOKEN_EXPIRATION):
        """
        Connect to the REST API.
        """
        self.user = user
        self.expires_in = expires_in
        self.expires_at = 0
        self.client = None

    def connect(self):
        """
        Connect to the REST API, authenticating with a JWT for the current user.
        """
        if JwtBuilder is None:
            raise NotConnectedToOpenEdX("This package must be installed in an OpenEdX environment.")

        now = int(time())
        scopes = ['profile', 'email']
        jwt = JwtBuilder(self.user).build_token(scopes, self.expires_in)
        self.client = EdxRestApiClient(
            self.API_BASE_URL, append_slash=self.APPEND_SLASH, jwt=jwt,
        )
        self.expires_at = now + self.expires_in

    def token_expired(self):
        """
        Return True if the JWT token has expired, False if not.
        """
        return int(time()) > self.expires_at

    @staticmethod
    def refresh_token(func):
        """
        Use this method decorator to ensure the JWT token is refreshed when needed.
        """
        @wraps(func)
        def inner(self, *args, **kwargs):
            """
            Before calling the wrapped function, we check if the JWT token is expired, and if so, re-connect.
            """
            if self.token_expired():
                self.connect()
            return func(self, *args, **kwargs)
        return inner


class EmbargoApiClient(object):
    """
    Client interface for using the edx-platform embargo API.
    """

    @staticmethod
    def redirect_if_blocked(course_run_ids, user=None, ip_address=None, url=None):
        """
        Return redirect to embargo error page if the given user is blocked.
        """
        for course_run_id in course_run_ids:
            redirect_url = embargo_api.redirect_if_blocked(
                CourseKey.from_string(course_run_id),
                user=user,
                ip_address=ip_address,
                url=url
            )
            if redirect_url:
                return redirect_url


class EnrollmentApiClient(LmsApiClient):
    """
    Object builds an API client to make calls to the Enrollment API.
    """

    API_BASE_URL = settings.ENTERPRISE_ENROLLMENT_API_URL

    def get_course_details(self, course_id):
        """
        Query the Enrollment API for the course details of the given course_id.

        Args:
            course_id (str): The string value of the course's unique identifier

        Returns:
            dict: A dictionary containing details about the course, in an enrollment context (allowed modes, etc.)
        """
        try:
            return self.client.course(course_id).get()
        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to retrieve course enrollment details for course [%s] due to: [%s]',
                course_id, str(exc)
            )
            return {}

    def _sort_course_modes(self, modes):
        """
        Sort the course mode dictionaries by slug according to the COURSE_MODE_SORT_ORDER constant.

        Arguments:
            modes (list): A list of course mode dictionaries.
        Returns:
            list: A list with the course modes dictionaries sorted by slug.

        """
        def slug_weight(mode):
            """
            Assign a weight to the course mode dictionary based on the position of its slug in the sorting list.
            """
            sorting_slugs = COURSE_MODE_SORT_ORDER
            sorting_slugs_size = len(sorting_slugs)
            if mode['slug'] in sorting_slugs:
                return sorting_slugs_size - sorting_slugs.index(mode['slug'])
            return 0
        # Sort slug weights in descending order
        return sorted(modes, key=slug_weight, reverse=True)

    def get_course_modes(self, course_id):
        """
        Query the Enrollment API for the specific course modes that are available for the given course_id.

        Arguments:
            course_id (str): The string value of the course's unique identifier

        Returns:
            list: A list of course mode dictionaries.

        """
        details = self.get_course_details(course_id)
        modes = details.get('course_modes', [])
        return self._sort_course_modes(modes)

    def has_course_mode(self, course_run_id, mode):
        """
        Query the Enrollment API to see whether a course run has a given course mode available.

        Arguments:
            course_run_id (str): The string value of the course run's unique identifier

        Returns:
            bool: Whether the course run has the given mode avaialble for enrollment.

        """
        course_modes = self.get_course_modes(course_run_id)
        return any(course_mode for course_mode in course_modes if course_mode['slug'] == mode)

    def enroll_user_in_course(self, username, course_id, mode):
        """
        Call the enrollment API to enroll the user in the course specified by course_id.

        Args:
            username (str): The username by which the user goes on the OpenEdX platform
            course_id (str): The string value of the course's unique identifier
            mode (str): The enrollment mode which should be used for the enrollment

        Returns:
            dict: A dictionary containing details of the enrollment, including course details, mode, username, etc.

        """
        return self.client.enrollment.post(
            {
                'user': username,
                'course_details': {'course_id': course_id},
                'mode': mode,
            }
        )

    def get_course_enrollment(self, username, course_id):
        """
        Query the enrollment API to get information about a single course enrollment.

        Args:
            username (str): The username by which the user goes on the OpenEdX platform
            course_id (str): The string value of the course's unique identifier

        Returns:
            dict: A dictionary containing details of the enrollment, including course details, mode, username, etc.

        """
        endpoint = getattr(
            self.client.enrollment,
            '{username},{course_id}'.format(username=username, course_id=course_id)
        )
        try:
            result = endpoint.get()
        except HttpNotFoundError:
            # This enrollment data endpoint returns a 404 if either the username or course_id specified isn't valid
            LOGGER.error(
                'Course enrollment details not found for invalid username or course; username=[%s], course=[%s]',
                username,
                course_id
            )
            return None
        # This enrollment data endpoint returns an empty string if the username and course_id is valid, but there's
        # no matching enrollment found
        if not result:
            LOGGER.info('Failed to find course enrollment details for user [%s] and course [%s]', username, course_id)
            return None

        return result

    def is_enrolled(self, username, course_run_id):
        """
        Query the enrollment API and determine if a learner is enrolled in a course run.

        Args:
            username (str): The username by which the user goes on the OpenEdX platform
            course_run_id (str): The string value of the course's unique identifier

        Returns:
            bool: Indicating whether the user is enrolled in the course run. Returns False under any errors.

        """
        enrollment = self.get_course_enrollment(username, course_run_id)
        return enrollment is not None and enrollment.get('is_active', False)

    def get_enrolled_courses(self, username):
        """
        Query the enrollment API to get a list of the courses a user is enrolled in.

        Args:
            username (str): The username by which the user goes on the OpenEdX platform

        Returns:
            list: A list of course objects, along with relevant user-specific enrollment details.

        """
        return self.client.enrollment.get(user=username)


class CourseApiClient(LmsApiClient):
    """
    Object builds an API client to make calls to the Course API.
    """

    API_BASE_URL = settings.LMS_INTERNAL_ROOT_URL + '/api/courses/v1/'
    APPEND_SLASH = True

    def get_course_details(self, course_id):
        """
        Retrieve all available details about a course.

        Args:
            course_id (str): The course ID identifying the course for which to retrieve details.

        Returns:
            dict: Contains keys identifying those course details available from the courses API (e.g., name).
        """
        try:
            return self.client.courses(course_id).get()
        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            LOGGER.exception('Details not found for course [%s] due to: [%s]', course_id, str(exc))
            return None


class ThirdPartyAuthApiClient(LmsApiClient):
    """
    Object builds an API client to make calls to the Third Party Auth API.
    """

    API_BASE_URL = settings.LMS_INTERNAL_ROOT_URL + '/api/third_party_auth/v0/'

    def get_remote_id(self, identity_provider, username):
        """
        Retrieve the remote identifier for the given username.

        Args:
        * ``identity_provider`` (str): identifier slug for the third-party authentication service used during SSO.
        * ``username`` (str): The username ID identifying the user for which to retrieve the remote name.

        Returns:
            string or None: the remote name of the given user.  None if not found.
        """
        return self._get_results(identity_provider, 'username', username, 'remote_id')

    def get_username_from_remote_id(self, identity_provider, remote_id):
        """
        Retrieve the remote identifier for the given username.

        Args:
        * ``identity_provider`` (str): identifier slug for the third-party authentication service used during SSO.
        * ``remote_id`` (str): The remote id identifying the user for which to retrieve the usernamename.

        Returns:
            string or None: the username of the given user.  None if not found.
        """
        return self._get_results(identity_provider, 'remote_id', remote_id, 'username')

    def _get_results(self, identity_provider, param_name, param_value, result_field_name):
        """
        Calls the third party auth api endpoint to get the mapping between usernames and remote ids.
        """
        try:
            kwargs = {param_name: param_value}
            returned = self.client.providers(identity_provider).users.get(**kwargs)
            results = returned.get('results', [])
        except HttpNotFoundError:
            LOGGER.error(
                'username not found for third party provider={provider}, {querystring_param}={id}'.format(
                    provider=identity_provider,
                    querystring_param=param_name,
                    id=param_value
                )
            )
            results = []

        for row in results:
            if row.get(param_name) == param_value:
                return row.get(result_field_name)
        return None


class GradesApiClient(JwtLmsApiClient):
    """
    Object builds an API client to make calls to the LMS Grades API.

    Note that this API client requires a JWT token, and so it keeps its token alive.
    """

    API_BASE_URL = settings.LMS_INTERNAL_ROOT_URL + '/api/grades/v0/'
    APPEND_SLASH = True

    @JwtLmsApiClient.refresh_token
    def get_course_grade(self, course_id, username):
        """
        Retrieve the grade for the given username for the given course_id.

        Args:
        * ``course_id`` (str): The string value of the course's unique identifier
        * ``username`` (str): The username ID identifying the user for which to retrieve the grade.

        Raises:

        HttpNotFoundError if no grade found for the given user+course.

        Returns:

        a dict containing:

        * ``username``: A string representation of a user's username passed in the request.
        * ``course_key``: A string representation of a Course ID.
        * ``passed``: Boolean representing whether the course has been passed according the course's grading policy.
        * ``percent``: A float representing the overall grade for the course
        * ``letter_grade``: A letter grade as defined in grading_policy (e.g. 'A' 'B' 'C' for 6.002x) or None

        """
        results = self.client.course_grade(course_id).users().get(username=username)
        for row in results:
            if row.get('username') == username:
                return row

        raise HttpNotFoundError('No grade record found for course={}, username={}'.format(course_id, username))


class CertificatesApiClient(JwtLmsApiClient):
    """
    Object builds an API client to make calls to the LMS Certificates API.

    Note that this API client requires a JWT token, and so it keeps its token alive.
    """

    API_BASE_URL = settings.LMS_INTERNAL_ROOT_URL + '/api/certificates/v0/'
    APPEND_SLASH = True

    @JwtLmsApiClient.refresh_token
    def get_course_certificate(self, course_id, username):
        """
        Retrieve the certificate for the given username for the given course_id.

        Args:
        * ``course_id`` (str): The string value of the course's unique identifier
        * ``username`` (str): The username ID identifying the user for which to retrieve the certificate

        Raises:

        HttpNotFoundError if no certificate found for the given user+course.

        Returns:

        a dict containing:

        * ``username``: A string representation of an user's username passed in the request.
        * ``course_id``: A string representation of a Course ID.
        * ``certificate_type``: A string representation of the certificate type.
        * ``created_date`: Datetime the certificate was created (tz-aware).
        * ``status``: A string representation of the certificate status.
        * ``is_passing``: True if the certificate has a passing status, False if not.
        * ``download_url``: A string representation of the certificate url.
        * ``grade``: A string representation of a float for the user's course grade.

        """
        return self.client.certificates(username).courses(course_id).get()


def enroll_user_in_course_locally(user, course_id, mode):
    """
    Enroll a user in a course, using local database methods.

    This is only used in one place; in models.py. It is necessary
    because the enrollment API can't enroll a user before it exists, and
    the post_save signal for users results in a user not actually being
    saved until after the API call is made.

    In Django 1.9 and later, there's a transaction.on_commit hook that we can
    use to create a callback. Once we're able to depend on having Django 1.9, we
    can shift over to that, but for right now, we have to do it this way.
    """
    if CourseEnrollment is None:
        raise NotConnectedToOpenEdX("This package must be installed in an OpenEdX environment.")

    CourseEnrollment.enroll(user, CourseKey.from_string(course_id), mode=mode, check_access=True)


def parse_lms_api_datetime(datetime_string, datetime_format=LMS_API_DATETIME_FORMAT):
    """
    Parse a received datetime into a timezone-aware, Python datetime object.

    Arguments:
        datetime_string: A string to be parsed.
        datetime_format: A datetime format string to be used for parsing

    """
    if isinstance(datetime_string, datetime.datetime):
        date_time = datetime_string
    else:
        try:
            date_time = datetime.datetime.strptime(datetime_string, datetime_format)
        except ValueError:
            date_time = datetime.datetime.strptime(datetime_string, LMS_API_DATETIME_FORMAT_WITHOUT_TIMEZONE)

    # If the datetime format didn't include a timezone, then set to UTC.
    # Note that if we're using the default LMS_API_DATETIME_FORMAT, it ends in 'Z',
    # which denotes UTC for ISO-8661.
    if date_time.tzinfo is None:
        date_time = date_time.replace(tzinfo=timezone.utc)
    return date_time



# -*- coding: utf-8 -*-
"""
Course metadata exporter for Enterprise Integrated Channel Degreed.
"""

from __future__ import absolute_import, unicode_literals

import json
from logging import getLogger

from integrated_channels.integrated_channel.exporters.course_metadata import CourseExporter

from django.conf import settings

from enterprise.utils import is_course_run_enrollable

LOGGER = getLogger(__name__)


class DegreedCourseExporter(CourseExporter):  # pylint: disable=abstract-method
    """
    Class to provide data transforms for Degreed course metadata export task.
    """

    CHUNK_PAGE_LENGTH = 1000
    SHORT_STRING_LIMIT = 255
    LONG_STRING_LIMIT = 2000

    STATUS_ACTIVE = 'ACTIVE'
    STATUS_INACTIVE = 'INACTIVE'

    def export(self):
        """
        Return serialized blocks of course metadata, which may be POST'd or DELETE'd depending on course enrollability.

        Yields:
            tuple: Contains the serialized course metadata JSON dump, and the method to be used.
                    - Method is one of 'POST' or 'DELETE'.
        """
        unenrollable_courses = []
        enrollable_courses = []
        for course_run in self.courses:
            # Course runs that have more than just the `contentId` key are considered enrollable.
            if len(course_run) > 1:
                enrollable_courses.append(course_run)
            else:
                unenrollable_courses.append(course_run)
        for course_group, method in [(unenrollable_courses, 'DELETE'), (enrollable_courses, 'POST')]:
            if course_group:
                yield (
                    json.dumps(
                        {
                            'courses': course_group,
                            'orgCode': self.enterprise_configuration.degreed_company_id,
                            'providerCode': self.enterprise_configuration.provider_id,
                        },
                        sort_keys=True
                    ).encode('utf-8'),
                    method
                )

    @property
    def data(self):
        """
        Return a transformed set of data ready to be exported.
        """
        return {
            'contentId': self.transform_course_id,
            'authors': self.transform_authors,
            'categoryTags': self.transform_category_tags,
            'url': self.transform_url,
            'imageUrl': self.transform_image_url,
            'videoUrl': self.transform_video_url,
            'title': self.transform_title,
            'description': self.transform_description,
            'difficulty': self.transform_difficulty,
            'duration': self.transform_duration,
            'publishDate': self.transform_publish_date,
            'format': self.transform_format,
            'institution': self.transform_institution,
            'costType': self.transform_cost_type,
            'language': self.transform_language_code
        }

    def transform(self, course_run):
        """
        Parse the provided course into the format natively supported by the provider.

        For Degreed, if a course run is unenrollable, it shouldn't be appearing in the
        upstream catalog. In that case, we just prepare a payload that is meant for deletion.
        The transmitter will do a check itself to see enrollability, and properly ``DELETE``
        the course.
        """
        LOGGER.info('Processing course run with ID [%s]', course_run['key'])
        LOGGER.debug('Parsing course run for [%s]: [%s]', self.enterprise_customer, json.dumps(course_run, indent=4))

        # Add the enterprise customer to the course run details so it can be used in the data transform
        course_run['enterprise_customer'] = self.enterprise_customer
        transformed_data = {}
        if is_course_run_enrollable(course_run):
            for key, transform in self.data.items():
                transformed_data[key] = transform(course_run) if transform is not None else course_run.get(key)
        else:
            # For courses that should be deleted, we just need their ID.
            transformed_data['contentId'] = self.transform_course_id(course_run)
        return transformed_data

    def transform_category_tags(self, course_run):  # pylint: disable=unused-argument
        """
        Return a transformed version of course tags.

        Note: no formal category tags for courses actually exist at this time. Update this when and if they do.
        """
        return course_run.get('tags') or []

    def transform_duration(self, course_run):
        """
        Return the transformed duration of the course.

        Degreed expects the value to be in minutes.
        """
        weeks_to_complete = course_run.get('weeks_to_complete') or 0
        return weeks_to_complete * 7 * 24 * 60

    def transform_publish_date(self, course_run):
        """
        Return the transformed version of the publish date for this course run.

        Example:
            2017-02-01T05:00:00Z -> 2017-02-01
        """
        # Instead of doing something fancy, since the first 10 characters will always be
        # of format yyyy-mm-dd, just get the first 10 characters.
        start = course_run.get('start') or ''
        return start[:10]

    def transform_institution(self, course_run):
        """
        Return the transformed version of the course institution.
        """
        institutions = course_run.get('owners')
        return institutions[0]['name'] if institutions else ''

    def transform_authors(self, course_run):
        """
        Return the array of author/instructor names for the course.
        """
        return [
            '{first_name} {last_name}'.format(
                first_name=instructor.get('given_name') or '',
                last_name=instructor.get('family_name') or '',
            ) for instructor in course_run.get('staff') or []
        ]

    def transform_language_code(self, course_run):
        """
        Return the ISO 639-1 language code that Degreed expects.

        Example:
            en-us -> en
            None -> en
        """
        code = course_run.get('content_language') or ''
        return code.split('-')[0] or 'en'

    def transform_course_id(self, course_run):
        """
        Return the transformed version of the course ID.
        """
        return course_run['key']

    def transform_cost_type(self, course_run):
        """
        Return the transformed version of the course enrollment cost type.

        One can use the course track to determine payment status. Values must be one of:
            - Subscription
            - Free
            - Paid

        We only really use the latter 2.
        """
        audit_modes = getattr(settings, 'ENTERPRISE_COURSE_ENROLLMENT_AUDIT_MODES', ['audit', 'honor'])
        return 'Free' if course_run.get('type') in audit_modes else 'Paid'

    def transform_title(self, course_run):
        """
        Return the transformed version of the course title.
        """
        return self.format_title(course_run)

    def transform_description(self, course_run):
        """
        Return the transformed version of the course description.

        We choose one value out of the course's full description, short description, and title
        depending on availability and length limits.
        """
        full_description = course_run.get('full_description') or ''
        if 0 < len(full_description) <= self.LONG_STRING_LIMIT:  # pylint: disable=len-as-condition
            return full_description
        return course_run.get('short_description') or course_run.get('title') or ''

    def transform_url(self, course_run):
        """
        Return the transformed version of the course's track selection URL.
        """
        return (
            course_run.get('enrollment_url')
            or course_run['enterprise_customer'].get_course_run_enrollment_url(course_run['key'])
        )

    def transform_image_url(self, course_run):
        """
        Return the transformed version of the course's image URL.
        """
        image = course_run.get('image') or {}
        return image.get('src') or ''

    def transform_video_url(self, course_run):
        """
        Return the transformed version of the course's video URL.
        """
        video = course_run.get('video') or {}
        return video.get('src') or ''

    def transform_difficulty(self, course_run):
        """
        Return the transformed difficulty.
        """
        return course_run.get('level_type') or ''

    def transform_format(self, course_run):
        """
        Return the transformed version of the course run format.

        Degreed expects one of the following values:
            - Instructor
            - Online
            - Virtual
            - Accredited
        """
        return 'Instructor' if course_run.get('pacing_type') == 'instructor_paced' else 'Online'



# -*- coding: utf-8 -*-
"""
Package for generic data exporters which serialize data to be transmitted to integrated channels.
"""

from __future__ import absolute_import, unicode_literals


class Exporter(object):
    """
    Interface for exporting data to be transmitted to an integrated channel.

    The interface contains the following method(s):

    export()
        Yields a serialized piece of data plus the HTTP method to be used by the transmitter.
    """

    def __init__(self, user, enterprise_configuration):
        """
        Store the data needed to export the learner data to the integrated channel.

        Arguments:
            * user: User instance with access to the Grades API for the Enterprise Customer's courses.
            * enterprise_configuration - The configuration connecting an enterprise to an integrated channel.
        """
        self.user = user
        self.enterprise_configuration = enterprise_configuration
        self.enterprise_customer = enterprise_configuration.enterprise_customer

    def export(self):
        """
        Export (read: serialize) data to be used by a transmitter to transmit to an integrated channel.
        """
        raise NotImplementedError('Implement in concrete subclass transmitter.')



"""
Celery tasks for integrated channel management commands.
"""

from __future__ import absolute_import, unicode_literals

import time

from celery import shared_task
from celery.utils.log import get_task_logger
from integrated_channels.integrated_channel.management.commands import INTEGRATED_CHANNEL_CHOICES

from django.contrib.auth.models import User

LOGGER = get_task_logger(__name__)


@shared_task
def transmit_course_metadata(username, channel_code, channel_pk):
    """
    Task to send course metadata to each linked integrated channel.

    Arguments:
        username (str): The username of the User to be used for making API requests for course metadata.
        channel_code (str): Capitalized identifier for the integrated channel
        channel_pk (str): Primary key for identifying integrated channel

    """
    start = time.time()
    api_user = User.objects.get(username=username)
    integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel_code].objects.get(pk=channel_pk)
    LOGGER.info('Processing course runs for integrated channel using configuration: [%s]', integrated_channel)
    try:
        integrated_channel.transmit_course_data(api_user)
    except Exception:  # pylint: disable=broad-except
        LOGGER.exception(
            'Transmission of course metadata failed for user [%s] and for integrated '
            'channel with code [%s] and id [%s].', username, channel_code, channel_pk
        )
    duration = time.time() - start
    LOGGER.info(
        'Course metadata transmission task for integrated channel configuration [%s] took [%s] seconds',
        integrated_channel,
        duration
    )


@shared_task
def transmit_learner_data(username, channel_code, channel_pk):
    """
    Task to send learner data to each linked integrated channel.

    Arguments:
        username (str): The username of the User to be used for making API requests for learner data.
        channel_code (str): Capitalized identifier for the integrated channel
        channel_pk (str): Primary key for identifying integrated channel

    """
    start = time.time()
    api_user = User.objects.get(username=username)
    integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel_code].objects.get(pk=channel_pk)
    LOGGER.info('Processing learners for integrated channel using configuration: [%s]', integrated_channel)

    # Note: learner data transmission code paths don't raise any uncaught exception, so we don't need a broad
    # try-except block here.
    integrated_channel.transmit_learner_data(api_user)

    duration = time.time() - start
    LOGGER.info(
        'Learner data transmission task for integrated channel configuration [%s] took [%s] seconds',
        integrated_channel,
        duration
    )



# -*- coding: utf-8 -*-
"""
Utilities to get details from the course catalog API.
"""
from __future__ import absolute_import, unicode_literals

from logging import getLogger

from edx_rest_api_client.client import EdxRestApiClient
from edx_rest_api_client.exceptions import SlumberBaseException
from opaque_keys import InvalidKeyError
from opaque_keys.edx.keys import CourseKey
from requests.exceptions import ConnectionError, Timeout  # pylint: disable=redefined-builtin

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _

from enterprise.utils import (
    MultipleProgramMatchError,
    NotConnectedToOpenEdX,
    get_configuration_value_for_site,
    get_course_id_from_course_run_id,
)

try:
    from openedx.core.lib.token_utils import JwtBuilder
except ImportError:
    JwtBuilder = None

try:
    from openedx.core.djangoapps.catalog.models import CatalogIntegration
except ImportError:
    CatalogIntegration = None

try:
    from openedx.core.lib.edx_api_utils import get_edx_api_data
except ImportError:
    get_edx_api_data = None


LOGGER = getLogger(__name__)


def course_discovery_api_client(user, catalog_url):
    """
    Return a Course Discovery API client setup with authentication for the specified user.
    """
    if JwtBuilder is None:
        raise NotConnectedToOpenEdX(
            _("To get a Catalog API client, this package must be "
              "installed in an Open edX environment.")
        )

    scopes = ['email', 'profile']
    expires_in = settings.OAUTH_ID_TOKEN_EXPIRATION
    jwt = JwtBuilder(user).build_token(scopes, expires_in)
    return EdxRestApiClient(catalog_url, jwt=jwt)


class CourseCatalogApiClient(object):
    """
    Object builds an API client to make calls to the Catalog API.
    """

    SEARCH_ALL_ENDPOINT = 'search/all/'
    CATALOGS_ENDPOINT = 'catalogs'
    CATALOGS_COURSES_ENDPOINT = 'catalogs/{}/courses/'
    COURSES_ENDPOINT = 'courses'
    COURSE_RUNS_ENDPOINT = 'course_runs'
    PROGRAMS_ENDPOINT = 'programs'
    PROGRAM_TYPES_ENDPOINT = 'program_types'

    DEFAULT_VALUE_SAFEGUARD = object()

    def __init__(self, user, site=None):
        """
        Create an Course Catalog API client setup with authentication for the specified user.

        This method retrieves an authenticated API client that can be used
        to access the course catalog API. It raises an exception to be caught at
        a higher level if the package doesn't have OpenEdX resources available.
        """
        if CatalogIntegration is None:
            raise NotConnectedToOpenEdX(
                _("To get a CatalogIntegration object, this package must be "
                  "installed in an Open edX environment.")
            )
        if get_edx_api_data is None:
            raise NotConnectedToOpenEdX(
                _("To parse a Catalog API response, this package must be "
                  "installed in an Open edX environment.")
            )

        self.user = user
        catalog_url = get_configuration_value_for_site(
            site,
            'COURSE_CATALOG_API_URL',
            settings.COURSE_CATALOG_API_URL
        )
        self.client = course_discovery_api_client(user, catalog_url)

    def get_search_results(self, querystring=None, traverse_pagination=True):
        """
        Return results from the discovery service's search/all endpoint.

        Arguments:
            querystring (dict): Querystring parameters used to filter search results.
            traverse_pagination (bool): True to return all results, False to return the paginated response.
                                        Defaults to True.
        Returns:
            list or dict: The paginated response dict if traverse_pagination is False, otherwise the extracted
                          results list.

        """
        return self._load_data(
            self.SEARCH_ALL_ENDPOINT,
            default=[],
            querystring=querystring,
            traverse_pagination=traverse_pagination,
            many=False,
        )

    def get_all_catalogs(self):
        """
        Return a list of all course catalogs, including name and ID.

        Returns:
            list: List of catalogs available for the user.

        """
        return self._load_data(
            self.CATALOGS_ENDPOINT,
            default=[]
        )

    def get_catalog(self, catalog_id):
        """
        Return specified course catalog.

        Returns:
            dict: catalog details if it is available for the user.

        """
        return self._load_data(
            self.CATALOGS_ENDPOINT,
            default=[],
            resource_id=catalog_id
        )

    def get_paginated_catalog_courses(self, catalog_id, querystring=None):
        """
        Return paginated response for all catalog courses.

        Returns:
            dict: API response with links to next and previous pages.

        """
        return self._load_data(
            self.CATALOGS_COURSES_ENDPOINT.format(catalog_id),
            default=[],
            querystring=querystring,
            traverse_pagination=False,
            many=False,
        )

    def get_paginated_catalogs(self, querystring=None):
        """
        Return a paginated list of course catalogs, including name and ID.

        Returns:
            dict: Paginated response containing catalogs available for the user.

        """
        return self._load_data(
            self.CATALOGS_ENDPOINT,
            default=[],
            querystring=querystring,
            traverse_pagination=False,
            many=False
        )

    def get_catalog_courses(self, catalog_id):
        """
        Return the courses included in a single course catalog by ID.

        Args:
            catalog_id (int): The catalog ID we want to retrieve.

        Returns:
            list: Courses of the catalog in question

        """
        return self._load_data(
            self.CATALOGS_COURSES_ENDPOINT.format(catalog_id),
            default=[]
        )

    def get_course_and_course_run(self, course_run_id):
        """
        Return the course and course run metadata for the given course run ID.

        Arguments:
            course_run_id (str): The course run ID.

        Returns:
            tuple: The course metadata and the course run metadata.
        """
        # Parse the course ID from the course run ID.
        course_id = get_course_id_from_course_run_id(course_run_id)
        # Retrieve the course metadata from the catalog service.
        course = self.get_course_details(course_id)

        course_run = None
        if course:
            # Find the specified course run.
            course_run = None
            course_runs = [course_run for course_run in course['course_runs'] if course_run['key'] == course_run_id]
            if course_runs:
                course_run = course_runs[0]

        return course, course_run

    def get_course_details(self, course_id):
        """
        Return the details of a single course by id - not a course run id.

        Args:
            course_id (str): The unique id for the course in question.

        Returns:
            dict: Details of the course in question.

        """
        return self._load_data(
            self.COURSES_ENDPOINT,
            resource_id=course_id,
            many=False
        )

    def get_course_run(self, course_run_id):
        """
        Return course_run data, including name, ID and seats.

        Args:
            course_run_id(string): Course run ID (aka Course Key) in string format.

        Returns:
            dict: Course run data provided by Course Catalog API.

        """
        return self._load_data(
            self.COURSE_RUNS_ENDPOINT,
            resource_id=course_run_id
        )

    def get_program_by_title(self, program_title):
        """
        Return single program by name, or None if not found.

        Arguments:
            program_title(string): Program title as seen by students and in Course Catalog Admin

        Returns:
            dict: Program data provided by Course Catalog API

        """
        all_programs = self._load_data(self.PROGRAMS_ENDPOINT, default=[])
        matching_programs = [program for program in all_programs if program.get('title') == program_title]
        if len(matching_programs) > 1:
            raise MultipleProgramMatchError(len(matching_programs))
        elif len(matching_programs) == 1:
            return matching_programs[0]
        else:
            return None

    def get_program_by_uuid(self, program_uuid):
        """
        Return single program by UUID, or None if not found.

        Arguments:
            program_uuid(string): Program UUID in string form

        Returns:
            dict: Program data provided by Course Catalog API

        """
        return self._load_data(
            self.PROGRAMS_ENDPOINT,
            resource_id=program_uuid,
            default=None
        )

    def get_program_course_keys(self, program_uuid):
        """
        Get a list of the course IDs (not course run IDs) contained in the program.

        Arguments:
            program_uuid (str): Program UUID in string form

        Returns:
            list(str): List of course keys in string form that are included in the program

        """
        program_details = self.get_program_by_uuid(program_uuid)
        if not program_details:
            return []
        return [course['key'] for course in program_details.get('courses', [])]

    def get_program_type_by_slug(self, slug):
        """
        Get a program type by its slug.

        Arguments:
            slug (str): The slug to identify the program type.

        Returns:
            dict: A program type object.

        """
        return self._load_data(
            self.PROGRAM_TYPES_ENDPOINT,
            resource_id=slug,
            default=None,
        )

    def get_common_course_modes(self, course_run_ids):
        """
        Find common course modes for a set of course runs.

        This function essentially returns an intersection of types of seats available
        for each course run.

        Arguments:
            course_run_ids(Iterable[str]): Target Course run IDs.

        Returns:
            set: course modes found in all given course runs

        Examples:
            # run1 has prof and audit, run 2 has the same
            get_common_course_modes(['course-v1:run1', 'course-v1:run2'])
            {'prof', 'audit'}

            # run1 has prof and audit, run 2 has only prof
            get_common_course_modes(['course-v1:run1', 'course-v1:run2'])
            {'prof'}

            # run1 has prof and audit, run 2 honor
            get_common_course_modes(['course-v1:run1', 'course-v1:run2'])
            {}

            # run1 has nothing, run2 has prof
            get_common_course_modes(['course-v1:run1', 'course-v1:run2'])
            {}

            # run1 has prof and audit, run 2 prof, run3 has audit
            get_common_course_modes(['course-v1:run1', 'course-v1:run2', 'course-v1:run3'])
            {}

            # run1 has nothing, run 2 prof, run3 has prof
            get_common_course_modes(['course-v1:run1', 'course-v1:run2', 'course-v1:run3'])
            {}

        """
        available_course_modes = None
        for course_run_id in course_run_ids:
            course_run = self.get_course_run(course_run_id) or {}
            course_run_modes = {seat.get('type') for seat in course_run.get('seats', [])}

            if available_course_modes is None:
                available_course_modes = course_run_modes
            else:
                available_course_modes &= course_run_modes

            if not available_course_modes:
                return available_course_modes

        return available_course_modes

    def is_course_in_catalog(self, catalog_id, course_id):
        """
        Determine if the given course or course run ID is contained in the catalog with the given ID.

        Args:
            catalog_id (int): The ID of the catalog
            course_id (str): The ID of the course or course run

        Returns:
            bool: Whether the course or course run is contained in the given catalog
        """
        try:
            # Determine if we have a course run ID, rather than a plain course ID
            course_run_id = str(CourseKey.from_string(course_id))
        except InvalidKeyError:
            course_run_id = None

        endpoint = self.client.catalogs(catalog_id).contains

        if course_run_id:
            resp = endpoint.get(course_run_id=course_run_id)
        else:
            resp = endpoint.get(course_id=course_id)

        return resp.get('courses', {}).get(course_id, False)

    def _load_data(self, resource, default=DEFAULT_VALUE_SAFEGUARD, **kwargs):
        """
        Load data from API client.

        Arguments:
            resource(string): type of resource to load
            default(any): value to return if API query returned empty result. Sensible values: [], {}, None etc.

        Returns:
            dict: Deserialized response from Course Catalog API

        """
        default_val = default if default != self.DEFAULT_VALUE_SAFEGUARD else {}
        try:
            return get_edx_api_data(
                api_config=CatalogIntegration.current(),
                resource=resource,
                api=self.client,
                **kwargs
            ) or default_val
        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to load data from resource [%s] with kwargs [%s] due to: [%s]',
                resource, kwargs, str(exc)
            )
            return default_val


class CourseCatalogApiServiceClient(CourseCatalogApiClient):
    """
    Catalog API client which uses the configured Catalog service user.
    """

    def __init__(self, site=None):
        """
        Create an Course Catalog API client setup with authentication for the
        configured catalog service user.
        """
        if CatalogIntegration is None:
            raise NotConnectedToOpenEdX(
                _("To get a CatalogIntegration object, this package must be "
                  "installed in an Open edX environment.")
            )

        catalog_integration = CatalogIntegration.current()
        if catalog_integration.enabled:
            try:
                user = catalog_integration.get_service_user()
                super(CourseCatalogApiServiceClient, self).__init__(user, site)
            except ObjectDoesNotExist:
                raise ImproperlyConfigured(_("The configured CatalogIntegration service user does not exist."))
        else:
            raise ImproperlyConfigured(_("There is no active CatalogIntegration."))

    @classmethod
    def program_exists(cls, program_uuid):
        """
        Get whether the program exists or not.
        """
        try:
            return bool(cls().get_program_by_uuid(program_uuid))
        except ImproperlyConfigured:
            return False



# -*- coding: utf-8 -*-
"""
Django admin integration for the Consent application.
"""

from __future__ import absolute_import, unicode_literals

from django.contrib import admin
from simple_history.admin import SimpleHistoryAdmin

from consent.models import DataSharingConsent


@admin.register(DataSharingConsent)
class DataSharingConsentAdmin(SimpleHistoryAdmin):
    """
    Django admin model for PendingEnrollment
    """

    class Meta(object):
        """
        Meta class for ``DataSharingConsentAdmin``.
        """

        model = DataSharingConsent

    readonly_fields = (
        'enterprise_customer',
        'username',
        'course_id',
        'granted',
        'exists',
    )

    list_display = (
        'enterprise_customer',
        'username',
        'course_id',
        'granted',
        'exists',
    )

    ordering = (
        "username",
    )

    search_fields = (
        'enterprise_customer__name',
        'enterprise_customer__uuid',
        'username',
        'course_id',
    )

    def has_add_permission(self, request):
        """
        Disable add permission for DataSharingConsent.
        """
        return False

    def has_delete_permission(self, request, obj=None):
        """
        Disable deletion permission for DataSharingConsent.
        """
        return False



# -*- coding: utf-8 -*-
"""
Generic course metadata transmitter for integrated channels.
"""

from __future__ import absolute_import, unicode_literals

import json
import logging

from integrated_channels.integrated_channel.client import IntegratedChannelApiClient
from integrated_channels.integrated_channel.transmitters import Transmitter
from requests import RequestException

from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist

LOGGER = logging.getLogger(__name__)


class CourseTransmitter(Transmitter):
    """
    A generic course metadata transmitter.

    It may be subclassed by specific integrated channel course metadata transmitters for
    each integrated channel's particular course metadata transmission requirements and expectations.
    """

    def __init__(self, enterprise_configuration, client=IntegratedChannelApiClient):
        """
        By default, use the abstract integrated channel API client which raises an error when used if not subclassed.
        """
        super(CourseTransmitter, self).__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )

    def transmit(self, payload, **kwargs):  # pylint: disable=unused-argument
        """
        Transmit the course metadata payload to the integrated channel.
        """
        total_transmitted = 0
        errors = []
        status_codes = []
        for course_metadata, method in payload.export():
            status_code, body = self.transmit_block(course_metadata, method=method)
            status_codes.append(str(status_code))
            error_message = body if status_code >= 400 else ''
            if error_message:
                errors.append(error_message)
            else:
                total_transmitted += len(course_metadata)

        error_message = ', '.join(errors) if errors else ''
        code_string = ', '.join(status_codes)

        # pylint: disable=invalid-name
        CatalogTransmissionAudit = apps.get_model('integrated_channel', 'CatalogTransmissionAudit')
        try:
            last_catalog_transmission = CatalogTransmissionAudit.objects.filter(
                error_message='',
                enterprise_customer_uuid=self.enterprise_configuration.enterprise_customer.uuid,
                channel=self.enterprise_configuration.provider_id,
            ).latest('created')
            last_audit_summary = json.loads(last_catalog_transmission.audit_summary)
        except ObjectDoesNotExist:
            last_audit_summary = {}

        CatalogTransmissionAudit(
            enterprise_customer_uuid=self.enterprise_configuration.enterprise_customer.uuid,
            total_courses=len(payload.courses),
            status=code_string,
            error_message=error_message,
            audit_summary=json.dumps(payload.resolve_removed_courses(last_audit_summary)),
            channel=self.enterprise_configuration.provider_id,
        ).save()

    def transmit_block(self, course_metadata, method='POST'):
        """
        Transmit a block of course metadata to the integrated channel.

        Handle any request errors and logging requirements.

        Args:
            course_metadata (bytes): A set of bytes containing a page's worth of course metadata
            method (str): One of ``POST`` or ``DELETE``, deciding which action the client should take
                          for this course run.

        Returns:
            status_code (int): An integer status for the HTTP request
            body (str): The server's response body
        """
        LOGGER.info(course_metadata)

        if method == 'POST':
            method = self.client.create_course_content
        elif method == 'DELETE':
            method = self.client.delete_course_content
        else:
            raise ValueError('Invalid method provided for the transmission of this course block: {}'.format(method))

        try:
            status_code, body = method(course_metadata)
        except RequestException as request_exception:
            status_code = 500
            body = str(request_exception)

        if status_code >= 400:
            LOGGER.error('Failed to send course metadata for Enterprise Customer {}\nError Message {}'.
                         format(self.enterprise_configuration.enterprise_customer.name, body))
        else:
            LOGGER.debug('Successfully sent course metadata for Enterprise Customer {}'.
                         format(self.enterprise_configuration.enterprise_customer.name))

        return status_code, body



# -*- coding: utf-8 -*-
"""
Class for transmitting course metadata to Degreed.
"""

from __future__ import absolute_import, unicode_literals

from integrated_channels.degreed.client import DegreedAPIClient
from integrated_channels.integrated_channel.transmitters.course_metadata import CourseTransmitter


class DegreedCourseTransmitter(CourseTransmitter):
    """
    This transmitter transmits a course metadata export to Degreed.
    """

    def __init__(self, enterprise_configuration, client=DegreedAPIClient):
        """
        By default, use the ``DegreedAPIClient`` for course metadata transmission to Degreed.
        """
        super(DegreedCourseTransmitter, self).__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )



# -*- coding: utf-8 -*-
"""
Class for transmitting course metadata to SuccessFactors.
"""

from __future__ import absolute_import, unicode_literals

from integrated_channels.integrated_channel.transmitters.course_metadata import CourseTransmitter
from integrated_channels.sap_success_factors.client import SAPSuccessFactorsAPIClient


class SapSuccessFactorsCourseTransmitter(CourseTransmitter):
    """
    This transmitter transmits a course metadata export to SAPSF.
    """

    def __init__(self, enterprise_configuration, client=SAPSuccessFactorsAPIClient):
        """
        By default, use the ``SAPSuccessFactorsAPIClient`` for course metadata transmission to SAPSF.
        """
        super(SapSuccessFactorsCourseTransmitter, self).__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )



# -*- coding: utf-8 -*-
"""
Course metadata exporter for Enterprise Integrated Channel SAP SuccessFactors.
"""

from __future__ import absolute_import, unicode_literals

import json
import os
from logging import getLogger

from integrated_channels.integrated_channel.exporters.course_metadata import CourseExporter
from integrated_channels.sap_success_factors.constants import SUCCESSFACTORS_OCN_LANGUAGE_CODES
from integrated_channels.utils import (
    UNIX_MAX_DATE_STRING,
    UNIX_MIN_DATE_STRING,
    current_time_is_in_interval,
    parse_datetime_to_epoch_millis,
)

LOGGER = getLogger(__name__)
COURSE_URL_SCHEME = os.environ.get('SUCCESSFACTORS_COURSE_EXPORT_DEFAULT_URL_SCHEME', 'https')


class SapSuccessFactorsCourseExporter(CourseExporter):  # pylint: disable=abstract-method
    """
    Class to provide data transforms for SAP SuccessFactors course metadata export task.
    """

    CHUNK_PAGE_LENGTH = 1000
    STATUS_ACTIVE = 'ACTIVE'
    STATUS_INACTIVE = 'INACTIVE'

    def export(self):
        """
        Return serialized blocks of data representing the courses to be POSTed, 1000 at a time.

        Yields:
            bytes: JSON-serialized course metadata structure
        """
        location = 0
        total_course_count = len(self.courses)
        while (location < total_course_count) or not location:
            this_batch = self.courses[location:location+self.CHUNK_PAGE_LENGTH]
            location += self.CHUNK_PAGE_LENGTH
            yield json.dumps({'ocnCourses': this_batch}, sort_keys=True).encode('utf-8'), 'POST'

    @property
    def data(self):
        """
        Return a transformed set of data ready to be exported.
        """
        return {
            'courseID': self.transform_course_id,
            'providerID': self.transform_provider_id,
            'status': self.transform_status,
            'title': self.transform_title,
            'description': self.transform_description,
            'thumbnailURI': self.transform_thumbnail_uri,
            'content': self.transform_content,
            'price': self.transform_price,
            'schedule': self.transform_schedule,
            'revisionNumber': self.transform_revision_number,
        }

    def transform_language_code(self, code):
        """
        Transform an ISO language code (for example, en-us) into a language name in the format read by SuccessFactors.
        """
        if code is None:
            return 'English'

        components = code.split('-')
        language_code = components[0]
        if len(components) == 2:
            country_code = components[1]
        elif len(components) == 1:
            country_code = '_'
        else:
            raise ValueError('Language codes may only have up to two components. Could not parse: {}'.format(code))

        if language_code not in SUCCESSFACTORS_OCN_LANGUAGE_CODES:
            LOGGER.warning('Language [%s] is not supported by SAPSF. Sending English by default.', language_code)
            return 'English'

        language_family = SUCCESSFACTORS_OCN_LANGUAGE_CODES[language_code]
        language_name = language_family.get(country_code, language_family['_'])
        return language_name

    def transform_course_id(self, course_run):
        """
        Return the transformed version of the course ID.
        """
        return course_run['key']

    def transform_provider_id(self, course_run):  # pylint: disable=unused-argument
        """
        Return the transformed version of the provider ID.
        """
        return self.enterprise_configuration.provider_id

    def transform_status(self, course_run):
        """
        Return the transformed version of the course status.
        """
        return (
            self.STATUS_ACTIVE
            if course_run['availability'] in [self.AVAILABILITY_CURRENT, self.AVAILABILITY_UPCOMING]
            else self.STATUS_INACTIVE
        )

    def transform_title(self, course_run):
        """
        Return the transformed version of the course title, as well as the locale
        """
        return [{
            'locale': self.transform_language_code(course_run.get('content_language')),
            'value':  self.format_title(course_run)
        }]

    def transform_description(self, course_run):
        """
        Return the transformed version of the course description, as well as the locale.
        """
        return [{
            'locale': self.transform_language_code(course_run.get('content_language')),
            'value': (
                (course_run.get('full_description') or '')
                or (course_run.get('short_description') or '')
                or (course_run.get('title') or '')
            )
        }]

    def transform_thumbnail_uri(self, course_run):
        """
        Return the transformed version of the course's thumbnail URI.
        """
        image = course_run.get('image') or {}
        return image.get('src') or ''

    def transform_content(self, course_run):
        """
        Return the transformed version of the course content.

        Note that this is not actually course *content* per say, but rather just metadata.
        SAPSF expects the "content" key for this metadata, however, so we name the function
        "transform_content" for consistency.
        """
        return [{
            'providerID': self.enterprise_configuration.provider_id,
            'launchURL': self.get_launch_url(
                course_run['enterprise_customer'],
                course_run['key'],
                course_run.get('enrollment_url') or ''
            ),
            'contentTitle': self.format_title(course_run),
            'contentID': course_run['key'],
            'launchType': 3,
            'mobileEnabled': course_run.get('mobile_available', 'false')
        }]

    def transform_price(self, course_run):  # pylint: disable=unused-argument
        """
        Return the transformed version of the course price.
        """
        return []

    def transform_schedule(self, course_run):
        """
        Return the transformed version of the course schedule.
        """
        start = course_run.get('start') or UNIX_MIN_DATE_STRING
        end = course_run.get('end') or UNIX_MAX_DATE_STRING
        return [{
            'startDate': parse_datetime_to_epoch_millis(start),
            'endDate': parse_datetime_to_epoch_millis(end),
            'active': current_time_is_in_interval(start, end)
        }]

    def transform_revision_number(self, course_run):  # pylint: disable=unused-argument
        """
        Return the transformed version of the course revision number.
        """
        return 1

    def resolve_removed_courses(self, previous_audit_summary):  # pylint: disable=arguments-differ
        """
        Ensures courses that are no longer in the catalog get properly marked as inactive.

        Args:
            previous_audit_summary (dict): The previous audit summary from the last course export.

        Returns:
            An audit summary of courses with information about their presence in the catalog and current status.
        """
        if self.removed_courses_resolved:
            return {}

        new_audit_summary = {}
        new_courses = []

        for course in self.courses:
            course_key = course['courseID']
            course_status = course['status']

            # Remove the key from previous audit summary so we can process courses that are no longer present,
            # and keep course records for all previously pushed courses and new, active courses.
            if previous_audit_summary.pop(course_key, None) or course_status == self.STATUS_ACTIVE:
                new_courses.append(course)
                new_audit_summary[course_key] = {
                    'in_catalog': True,
                    'status': course_status,
                }

        for course_key, summary in previous_audit_summary.items():
            # Add a course payload to new_courses so that courses no longer in the catalog are marked inactive.
            if summary['status'] == self.STATUS_ACTIVE and summary['in_catalog']:
                new_courses.append(self.get_course_metadata_for_inactivation(
                    course_key,
                    self.enterprise_customer,
                    self.enterprise_configuration.provider_id
                ))

                new_audit_summary[course_key] = {
                    'in_catalog': False,
                    'status': self.STATUS_INACTIVE,
                }

        self.courses = new_courses
        self.removed_courses_resolved = True
        return new_audit_summary

    def get_launch_url(self, enterprise_customer, course_id, enrollment_url=None):
        """
        Given an EnterpriseCustomer and a course ID, determine the appropriate launch url.

        Args:
            enterprise_customer (EnterpriseCustomer): The EnterpriseCustomer that a URL needs to be built for
            course_id (str): The string identifier of the course in question
            enrollment_url (str): Enterprise landing page url for the given course from enterprise courses API
        """
        return enrollment_url or enterprise_customer.get_course_run_enrollment_url(course_id)

    def get_course_metadata_for_inactivation(self, course_id, enterprise_customer, provider_id):
        """
        Provide the minimal course metadata structure for updating a course to be inactive.
        """
        return {
            'courseID': course_id,
            'providerID': provider_id,
            'status': self.STATUS_INACTIVE,
            'title': [
                {
                    'locale': self.transform_language_code(None),
                    'value': course_id
                },
            ],
            'content': [
                {
                    'providerID': provider_id,
                    'launchURL': self.get_launch_url(enterprise_customer, course_id),
                    'contentTitle': 'Course Description',
                    'launchType': 3,
                    'contentID': course_id,
                }
            ],
        }



# -*- coding: utf-8 -*-
"""
Assist integrated channels with retrieving learner completion data.

Module contains resources for integrated pipelines to retrieve all the
grade and completion data for enrollments belonging to a particular
enterprise customer.
"""

from __future__ import absolute_import, unicode_literals

from logging import getLogger

from consent.models import DataSharingConsent
from integrated_channels.integrated_channel.exporters import Exporter
from integrated_channels.utils import parse_datetime_to_epoch_millis
from slumber.exceptions import HttpNotFoundError

from django.apps import apps
from django.utils import timezone
from django.utils.dateparse import parse_datetime

from enterprise.api_client.lms import CertificatesApiClient, CourseApiClient, GradesApiClient
from enterprise.models import EnterpriseCourseEnrollment

LOGGER = getLogger(__name__)


class LearnerExporter(Exporter):
    """
    Base class for exporting learner completion data to integrated channels.
    """

    GRADE_PASSING = 'Pass'
    GRADE_FAILING = 'Fail'
    GRADE_INCOMPLETE = 'In Progress'

    def __init__(self, user, enterprise_configuration):
        """
        Store the data needed to export the learner data to the integrated channel.

        Arguments:

        * ``user``: User instance with access to the Grades API for the Enterprise Customer's courses.
        * ``enterprise_configuration``: EnterpriseCustomerPluginConfiguration instance for the current channel.
        """
        # The Grades API and Certificates API clients require an OAuth2 access token,
        # so cache the client to allow the token to be reused. Cache other clients for
        # general reuse.
        self.grades_api = None
        self.certificates_api = None
        self.course_api = None
        self.course_enrollment_api = None
        super(LearnerExporter, self).__init__(user, enterprise_configuration)

    @property
    def grade_passing(self):
        """
        Returns the string used for a passing grade.
        """
        return self.GRADE_PASSING

    @property
    def grade_failing(self):
        """
        Returns the string used for a failing grade.
        """
        return self.GRADE_FAILING

    @property
    def grade_incomplete(self):
        """
        Returns the string used for an incomplete course grade.
        """
        return self.GRADE_INCOMPLETE

    def export(self):
        """
        Collect learner data for the ``EnterpriseCustomer`` where data sharing consent is granted.

        Yields a learner data object for each enrollment, containing:

        * ``enterprise_enrollment``: ``EnterpriseCourseEnrollment`` object.
        * ``completed_date``: datetime instance containing the course/enrollment completion date; None if not complete.
          "Course completion" occurs for instructor-paced courses when course certificates are issued, and
          for self-paced courses, when the course end date is passed, or when the learner achieves a passing grade.
        * ``grade``: string grade recorded for the learner in the course.
        """
        # Fetch the consenting enrollment data, including the enterprise_customer_user.
        # Order by the course_id, to avoid fetching course API data more than we have to.
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__enterprise_customer=self.enterprise_customer,
        ).order_by('course_id')

        # Fetch course details from the Course API, and cache between calls.
        course_details = None
        for enterprise_enrollment in enrollment_queryset:

            course_id = enterprise_enrollment.course_id

            # Fetch course details from Courses API
            # pylint: disable=unsubscriptable-object
            if course_details is None or course_details['course_id'] != course_id:
                if self.course_api is None:
                    self.course_api = CourseApiClient()
                course_details = self.course_api.get_course_details(course_id)

            if course_details is None:
                # Course not found, so we have nothing to report.
                LOGGER.error("No course run details found for enrollment [%d]: [%s]",
                             enterprise_enrollment.pk, course_id)
                continue

            consent = DataSharingConsent.objects.proxied_get(
                username=enterprise_enrollment.enterprise_customer_user.username,
                course_id=enterprise_enrollment.course_id,
                enterprise_customer=enterprise_enrollment.enterprise_customer_user.enterprise_customer
            )

            if not consent.granted or enterprise_enrollment.audit_reporting_disabled:
                continue

            # For instructor-paced courses, let the certificate determine course completion
            if course_details.get('pacing') == 'instructor':
                completed_date, grade, is_passing = self._collect_certificate_data(enterprise_enrollment)

            # For self-paced courses, check the Grades API
            else:
                completed_date, grade, is_passing = self._collect_grades_data(enterprise_enrollment, course_details)

            record = self.get_learner_data_record(
                enterprise_enrollment=enterprise_enrollment,
                completed_date=completed_date,
                grade=grade,
                is_passing=is_passing,
            )
            if record:
                # There are some cases where we won't receive a record from the above
                # method; right now, that should only happen if we have an Enterprise-linked
                # user for the integrated channel, and transmission of that user's
                # data requires an upstream user identifier that we don't have (due to a
                # failure of SSO or similar). In such a case, `get_learner_data_record`
                # would return None, and we'd simply skip yielding it here.
                yield record

    def get_learner_data_record(self, enterprise_enrollment, completed_date=None, grade=None, is_passing=False):
        """
        Generate a learner data transmission audit with fields properly filled in.
        """
        # pylint: disable=invalid-name
        LearnerDataTransmissionAudit = apps.get_model('integrated_channel', 'LearnerDataTransmissionAudit')
        completed_timestamp = None
        course_completed = False
        if completed_date is not None:
            completed_timestamp = parse_datetime_to_epoch_millis(completed_date)
            course_completed = is_passing

        return LearnerDataTransmissionAudit(
            enterprise_course_enrollment_id=enterprise_enrollment.id,
            course_id=enterprise_enrollment.course_id,
            course_completed=course_completed,
            completed_timestamp=completed_timestamp,
            grade=grade,
        )

    def _collect_certificate_data(self, enterprise_enrollment):
        """
        Collect the learner completion data from the course certificate.

        Used for Instructor-paced courses.

        If no certificate is found, then returns the completed_date = None, grade = In Progress, on the idea that a
        certificate will eventually be generated.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect completion/grade data

        Returns:
            completed_date: Date the course was completed, this is None if course has not been completed.
            grade: Current grade in the course.
            is_passing: Boolean indicating if the grade is a passing grade or not.
        """

        if self.certificates_api is None:
            self.certificates_api = CertificatesApiClient(self.user)

        course_id = enterprise_enrollment.course_id
        username = enterprise_enrollment.enterprise_customer_user.user.username

        try:
            certificate = self.certificates_api.get_course_certificate(course_id, username)
            completed_date = certificate.get('created_date')
            if completed_date:
                completed_date = parse_datetime(completed_date)
            else:
                completed_date = timezone.now()

            # For consistency with _collect_grades_data, we only care about Pass/Fail grades. This could change.
            is_passing = certificate.get('is_passing')
            grade = self.grade_passing if is_passing else self.grade_failing

        except HttpNotFoundError:
            completed_date = None
            grade = self.grade_incomplete
            is_passing = False

        return completed_date, grade, is_passing

    def _collect_grades_data(self, enterprise_enrollment, course_details):
        """
        Collect the learner completion data from the Grades API.

        Used for self-paced courses.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect completion/grade data
            course_details (dict): the course details for the course in the enterprise enrollment record.

        Returns:
            completed_date: Date the course was completed, this is None if course has not been completed.
            grade: Current grade in the course.
            is_passing: Boolean indicating if the grade is a passing grade or not.
        """
        if self.grades_api is None:
            self.grades_api = GradesApiClient(self.user)

        course_id = enterprise_enrollment.course_id
        username = enterprise_enrollment.enterprise_customer_user.user.username

        try:
            grades_data = self.grades_api.get_course_grade(course_id, username)

        except HttpNotFoundError:
            # Grade not found, so we have nothing to report.
            LOGGER.error("No grades data found for [%d]: [%s], [%s]", enterprise_enrollment.pk, course_id, username)
            return None, None, None

        # Prepare to process the course end date and pass/fail grade
        course_end_date = course_details.get('end')
        if course_end_date is not None:
            course_end_date = parse_datetime(course_end_date)
        now = timezone.now()
        is_passing = grades_data.get('passed')

        # We can consider a course complete if:
        # * the course's end date has passed
        if course_end_date is not None and course_end_date < now:
            completed_date = course_end_date
            grade = self.grade_passing if is_passing else self.grade_failing

        # * Or, the learner has a passing grade (as of now)
        elif is_passing:
            completed_date = now
            grade = self.grade_passing

        # Otherwise, the course is still in progress
        else:
            completed_date = None
            grade = self.grade_incomplete

        return completed_date, grade, is_passing
